@@ -1,15 +1,17 @@
 __all__ = ["dispatch"]
 
-import argparse
+import dataclasses
 import functools
 import logging
+import pathlib
 import sys
 import typing
 
 from . import __version__, procs, venvs
+from ._argparse import PyEMArgumentParser
 from ._logging import configure_logging
 from .errs import Error
-from .projects import Project, ProjectNotFound
+from .projects import Project
 
 
 _ArgIter = typing.Iterator[str]
@@ -23,11 +25,16 @@ logger = logging.getLogger(__name__)
 
 def _iter_until_subcommand(iterator: _ArgIter) -> _ArgIter:
     next(iterator)  # Skip argv[0].
+    keep_next = False
     for arg in iterator:
         yield arg
-        if not arg.startswith("-"):  # Subcommand met.
+        if keep_next:
+            keep_next = False
+        elif not arg.startswith("-"):  # Subcommand met.
             return
-    yield ""
+        elif "=" not in arg:
+            keep_next = True
+    yield ""  # Stub subcommand if the arg list is completely empty.
 
 
 def _partition_argv(argv: _ArgList) -> typing.Tuple[_ArgList, str, _ArgList]:
@@ -42,31 +49,16 @@ _MISSING_PARSER_EPILOG = (
 )
 
 
-class PyEMArgumentParser(argparse.ArgumentParser):
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self.add_argument(
-            "--project",
-            help="alternative path marking the project root",
-            metavar="PATH",
-            default=None,
-        )
-
-    def add_subparsers(self, **kwargs):
-        # Revert to the basic parser class to avoid root arguments from being
-        # defined in subparsers.
-        if "parser_class" not in kwargs:
-            kwargs["parser_class"] = argparse.ArgumentParser
-        return super().add_subparsers(**kwargs)
-
-
 def _parse_missing(argv: _ArgList) -> _Options:
     parser = PyEMArgumentParser(epilog=_MISSING_PARSER_EPILOG)
     parser.add_argument(
         "--version", action="version", version=f"%(prog)s {__version__}"
     )
+    parser.add_argument("--spec", help="venv context to use", default=None)
     parser.add_argument("cmd", help="command to run")
-    parser.add_argument("arg", nargs="*", help="command argument")
+    parser.add_argument(
+        "arg", nargs="*", default=None, help="command argument"
+    )
     return parser.parse_args(argv)
 
 
@@ -84,16 +76,28 @@ def _parse_for_venv(argv: _ArgList) -> _Options:
         "add", description="Create a virtual environment for this project."
     )
     parser_add.add_argument("python", help="base interpreter to use")
+    parser_add.add_argument(
+        "--set",
+        "--use",
+        dest="activate",
+        action="store_true",
+        default=False,
+        help="Set the newly created virtual environment as default.",
+    )
     parser_add.set_defaults(func=venvs.add)
 
     parser_rm = subparsers.add_parser(
-        "remove", description="Remove a virtual environment from this project."
+        "remove",
+        aliases=["rm"],
+        description="Remove a virtual environment from this project.",
     )
     parser_rm.add_argument("spec", help="venv specifier")
     parser_rm.set_defaults(func=venvs.remove)
 
     parser_set = subparsers.add_parser(
-        "set", description="Set the project's default virtual environment."
+        "set",
+        aliases=["use"],
+        description="Set the project's default virtual environment.",
     )
     parser_set.add_argument("spec", help="venv specifier")
     parser_set.set_defaults(func=venvs.activate)
@@ -130,26 +134,64 @@ def _parse_args(argv: _ArgList) -> _Options:
         return _parse_for_venv(before_cmd + after_cmd)
     if not cmd:
         return _parse_missing(before_cmd)
+    if cmd == "pip":
+        # Special-case pip to invoke "python -m pip" instead.
+        after_cmd = ["-m", "pip", *after_cmd]
+        return _parse_for_bridge(before_cmd, "python", after_cmd)
     return _parse_for_bridge(before_cmd, cmd, after_cmd)
 
 
+@dataclasses.dataclass()
+class _ProjectNotFound(Exception):
+    start: pathlib.Path
+
+
+# Ways to specify a project root. The .venvs directory is checked before
+# pyproject.toml, so an outer .venvs directory has precedence over an inner
+# pyproject.toml. This is important to resolve nested projects, where we
+# usually want to reach environments set up in the outmost project, instead of
+# creating new ones for each inner project.
+
+_DiscoveryVariant = typing.Tuple[str, typing.Callable[[pathlib.Path], bool]]
+
+_DISCOVERY_VARIANTS: typing.List[_DiscoveryVariant] = [
+    (".venvs", pathlib.Path.is_dir),
+    ("pyproject.toml", pathlib.Path.is_file),
+]
+
+
+def _project_from_path(start: pathlib.Path) -> Project:
+    path = start.resolve()
+    if any(path.name == n and f(path) for n, f in _DISCOVERY_VARIANTS):
+        path = path.parent
+    if path.is_dir():
+        return Project(path)
+    raise _ProjectNotFound(start)
+
+
+def _project_from_discovery() -> Project:
+    start = pathlib.Path.cwd()
+    for marker, check in _DISCOVERY_VARIANTS:
+        for path in start.joinpath(marker).parents:
+            if check(path.joinpath(marker)):
+                return Project(path)
+    raise _ProjectNotFound(start)
+
+
 def dispatch(argv: typing.Optional[_ArgList]) -> int:
     configure_logging(logging.INFO)  # TODO: Make this configurable.
 
     if argv is None:
         argv = sys.argv
-
     opts = _parse_args(argv)
 
-    # If we specify an explicit path, we only want to search the specified
-    # directory (or the directory containing the specified file), i.e. up 0
-    # times. Otherwise we search indefinitely (up=None).
-    up = 0 if opts.project else None
-
     try:
-        project = Project.discover(opts.project, up=up)
-    except ProjectNotFound as e:
-        logger.error("No pyproject.toml found in %s", e.start)
+        if opts.project is not None:
+            project = _project_from_path(pathlib.Path(opts.project))
+        else:
+            project = _project_from_discovery()
+    except _ProjectNotFound as e:
+        logger.error("No valid project from %s", e.start)
         return Error.project_not_found
 
     return opts.func(project, opts)
